const resultCallback = (result) => result;

const typeOf = (value) => Object.prototype.toString.call(value).slice(8, -1).toLowerCase();

const isArray = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'array' || typeof value === 'object') &&
    Array.isArray(value), value, payload);

const isFunction = (value, callback = resultCallback, payload) => callback(typeof value === 'function' ||
    (typeOf(value) === 'function' && value instanceof Function)
    ? /class/.test(Function.prototype.toString.call(value).slice(0, 5)) ===
        false
    : false, value, payload);

const areDeterminer = (checkFn, ...values) => {
    return {
        every: (callback = resultCallback, payload) => callback(values.every((value) => checkFn(value)), values, payload),
        forEach: (forEachCallback, payload) => {
            isArray(values) &&
                isFunction(forEachCallback) &&
                values.forEach((value, index) => forEachCallback(checkFn(value), value, index, values, payload));
        },
        some: (callback = resultCallback, payload) => callback(isArray(values) ? values.some((value) => checkFn(value)) : false, values, payload),
    };
};

const isBigInt = (value, callback = resultCallback, payload) => callback(typeof value === 'bigint', value, payload);

const areBigInt = (...values) => areDeterminer(isBigInt, ...values);

const isBoolean = (value, callback = resultCallback, payload) => callback((typeof value === 'boolean' ||
    typeOf(value) === 'boolean' ||
    (typeof value === 'object' && value instanceof Boolean)) &&
    (value.valueOf() === true || value.valueOf() === false), value, payload);

const areBoolean = (...values) => areDeterminer(isBoolean, ...values);

const isDate = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'date' || typeof value === 'object') &&
    value instanceof Date === true &&
    !isNaN(value), value, payload);

const areDate = (...values) => areDeterminer(isDate, ...values);

const isDefined = (value, callback = resultCallback, payload) => callback(typeof value !== 'undefined' && value !== undefined, value, payload);

const areDefined = (...values) => areDeterminer(isDefined, ...values);

const isFalse = (value, callback = resultCallback, payload) => callback(isBoolean(value) ? value.valueOf() === false : false, value, payload);

const areFalse = (...values) => areDeterminer(isFalse, ...values);

const isNull = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'null' || typeof value === 'object') && value === null, value, payload);

const areNull = (...values) => areDeterminer(isNull, ...values);

const isNumber = (value, callback = resultCallback, payload) => callback((typeof value === 'number' ||
    ((typeOf(value) === 'number' || typeof value === 'object') &&
        value instanceof Number)) &&
    !Number.isNaN(value.valueOf()) &&
    Number.isFinite(value.valueOf()), value, payload);

const areNumber = (...values) => areDeterminer(isNumber, ...values);

const isRegExp = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'regexp' || typeof value === 'object') &&
    value instanceof RegExp, value, payload);

const areRegExp = (...values) => areDeterminer(isRegExp, ...values);

const isStringObject = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'string' || typeof value === 'object') &&
    value instanceof String, value, payload);

const isStringType = (value, callback = resultCallback, payload) => callback(typeof value === 'string', value, payload);

const isString = (value, callback = resultCallback, payload) => callback(isStringType(value) || isStringObject(value), value, payload);

const areString = (...values) => areDeterminer(isString, ...values);

const isSymbol = (value, callback = resultCallback, payload) => callback(typeof value === 'symbol', value, payload);

const areSymbol = (...values) => areDeterminer(isSymbol, ...values);

const isTrue = (value, callback = resultCallback, payload) => callback(isBoolean(value) ? value.valueOf() === true : false, value, payload);

const areTrue = (...values) => areDeterminer(isTrue, ...values);

const isUndefined = (value, callback = resultCallback, payload) => callback(typeof value === 'undefined', value, payload);

const areUndefined = (...values) => areDeterminer(isUndefined, ...values);

const are = Object.freeze({
    bigint: areBigInt,
    boolean: areBoolean,
    date: areDate,
    defined: areDefined,
    false: areFalse,
    null: areNull,
    number: areNumber,
    regexp: areRegExp,
    string: areString,
    symbol: areSymbol,
    true: areTrue,
    undefined: areUndefined
});

const guardArray = (value, callback, payload) => isArray(value, callback, payload);

const guardBigInt = (value, callback, payload) => isBigInt(value, callback, payload);

const guardBoolean = (value, callback, payload) => isBoolean(value, callback, payload);

const isClass = (value, callback = resultCallback, payload) => callback(typeof value === 'function' ||
    (typeOf(value) === 'function' && value instanceof Function)
    ? /class/.test(Function.prototype.toString.call(value).slice(0, 5))
    : false, value, payload);

const guardClass = (value, callback, payload) => isClass(value, callback, payload);

const guardDate = (value, callback, payload) => isDate(value, callback, payload);

const guardDefined = (value, callback, payload) => isDefined(value, callback, payload);

const guardFalse = (value, callback, payload) => isFalse(value, callback, payload);

const guardFunction = (value, callback, payload) => isFunction(value, callback, payload);

const isObject = (value, callback = resultCallback, payload) => callback((typeof value === 'object' || typeOf(value) === 'object') &&
    value instanceof Object, value, payload);

const isInstance = (value, constructor, callback = resultCallback, payload) => callback(isObject(value) &&
    typeof constructor === 'function' &&
    constructor instanceof Function
    ? value instanceof constructor
    : false, value, Object.assign(Object.assign({}, payload), { ctor: constructor }));

const guardInstance = (value, constructor, callback, payload) => isInstance(value, constructor, callback, payload);

const isNumberType = (value, callback = resultCallback, payload) => callback(typeof value === 'number' && Number.isFinite(value) && !Number.isNaN(value), value, payload);

const isKey = (value, callback = resultCallback, payload) => callback(isStringType(value) || isNumberType(value) || isSymbol(value), value, payload);

const guardKey = (value, callback, payload) => isKey(value, callback, payload);

const guardNull = (value, callback, payload) => isNull(value, callback, payload);

const guardNumber = (value, callback, payload) => isNumber(value, callback, payload);

const isNumberBetween = (value, min, max, callback = resultCallback, payload) => callback(isNumber(value)
    ? (isNumberType(min) ? value.valueOf() >= min : false) &&
        (isNumberType(max) ? value.valueOf() <= max : false)
    : false, value, Object.assign(Object.assign({}, payload), { min, max }));

const guardNumberBetween = (value, min, max, callback, payload) => isNumberBetween(value, min, max, callback, payload);

const guardObject = (value, callback, payload) => isObject(value, callback, payload);

const isObjectKey = (value, key, callback = resultCallback, payload) => callback(isObject(value) ? {}.hasOwnProperty.call(value, key) : false, value, Object.assign(Object.assign({}, payload), { key }));

const guardObjectKey = (value, key, callback, payload) => isObjectKey(value, key, callback, payload);

const isObjectKeyIn = (value, key, callback = resultCallback, payload) => callback(isObject(value) ? key in value : false, value, Object.assign(Object.assign({}, payload), { key }));

const guardObjectKeyIn = (value, key, callback, payload) => isObjectKeyIn(value, key, callback, payload);

const isObjectKeys = (value, keys, callback = resultCallback, payload) => callback(isObject(value) && isArray(keys)
    ? keys.every((key) => ({}.hasOwnProperty.call(value, key)))
    : false, value, Object.assign(Object.assign({}, payload), { keys }));

const guardObjectKeys = (value, keys, callback, payload) => isObjectKeys(value, keys, callback, payload);

const isObjectKeysIn = (value, keys, callback = resultCallback, payload) => callback(isObject(value) && isArray(keys) ? keys.every((k) => k in value) : false, value, Object.assign(Object.assign({}, payload), { keys }));

const guardObjectKeysIn = (value, keys, callback, payload) => isObjectKeysIn(value, keys, callback, payload);

const isObjectSomeKeys = (value, keys, callback = resultCallback, payload) => callback(isObject(value) && isArray(keys)
    ? keys.some((someKey) => isArray(someKey)
        ? someKey.every((everyKey) => ({}.hasOwnProperty.call(value, everyKey)))
        : {}.hasOwnProperty.call(value, someKey) === true)
    : false, value, Object.assign(Object.assign({}, payload), { keys }));

const guardObjectSomeKeys = (value, keys, callback, payload) => isObjectSomeKeys(value, keys, callback, payload);

const isBooleanType = (value, callback = resultCallback, payload) => callback(typeof value === 'boolean' && (value === true || value === false), value, payload);

const isPrimitive = (value, type, callback = resultCallback, payload) => isStringType(type)
    ? {
        bigint: isBigInt,
        boolean: isBooleanType,
        number: isNumberType,
        null: isNull,
        string: isStringType,
        symbol: isSymbol,
        undefined: isUndefined,
    }[type](value, callback, payload)
    : callback(isNull(value) ||
        (typeof value !== 'object' && typeof value !== 'function'), value, payload);

const guardPrimitive = (value, type, callback, payload) => isPrimitive(value, type, callback, payload);

const guardRegExp = (value, callback, payload) => isRegExp(value, callback, payload);

const guardString = (value, callback, payload) => isString(value, callback, payload);

const isStringIncludes = (value, includes, callback = resultCallback, payload) => callback(isString(value)
    ? isArray(includes)
        ? includes.every((include) => value.valueOf().includes(include))
        : false
    : false, value, Object.assign(Object.assign({}, payload), { includes }));

const guardStringIncludes = (value, includes, callback = resultCallback, payload) => isStringIncludes(value, includes, callback, payload);

const isStringIncludesSome = (value, includes, callback = resultCallback, payload) => callback(isString(value)
    ? isArray(includes)
        ? includes.some((include) => value.valueOf().includes(include))
        : false
    : false, value, Object.assign(Object.assign({}, payload), { includes }));

const guardStringIncludesSome = (value, includes, callback = resultCallback, payload) => isStringIncludesSome(value, includes, callback, payload);

const isBooleanObject = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'boolean' || typeof value === 'object') &&
    value instanceof Boolean &&
    (value.valueOf() === true || value.valueOf() === false), value, payload);

const isNumberObject = (value, callback = resultCallback, payload) => callback((typeOf(value) === 'number' || typeof value === 'object') &&
    value instanceof Number &&
    !Number.isNaN(value.valueOf()) &&
    Number.isFinite(value.valueOf()), value, payload);

const isStringLength = (value, length, callback = resultCallback, payload) => callback(isString(value) && isNumberType(length) && length > 0
    ? value.valueOf().length === length
    : false, value, Object.assign(Object.assign({}, payload), { length }));

const isStringLengthBetween = (value, min, max, callback = resultCallback, payload) => callback(isString(value)
    ? (isNumberType(min) && min >= 0
        ? value.valueOf().length >= min
        : false) &&
        (isNumberType(max) && max >= 0
            ? value.valueOf().length <= max
            : false)
    : false, value, Object.assign(Object.assign({}, payload), { min, max }));

const isNotNull = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'null' && value !== null, value, payload);

const isType = (value, type, callback = resultCallback, payload) => isStringType(type)
    ? {
        bigint: isBigInt,
        boolean: isBooleanType,
        function: isFunction,
        number: isNumberType,
        object: isObject,
        null: isNull,
        string: isStringType,
        symbol: isSymbol,
        undefined: isUndefined,
    }[type](value, callback, payload)
    : isNotNull(type)
        ? isInstance(value, type, callback, payload)
        : false;

const isNotBoolean = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'boolean' &&
    typeof value !== 'boolean' &&
    value instanceof Boolean === false, value, payload);

const isNotDefined = (value, callback = resultCallback, payload) => callback(typeOf(value) === 'undefined' &&
    typeof value === 'undefined' &&
    value === undefined, value, payload);

const isNotFunction = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'function' &&
    typeof value !== 'function' &&
    value instanceof Function === false, value, payload);

const isNotNumber = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'number' &&
    typeof value !== 'number' &&
    value instanceof Number === false, value, payload);

const isNotString = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'string' &&
    typeof value !== 'string' &&
    value instanceof String === false, value, payload);

const isNotUndefined = (value, callback = resultCallback, payload) => callback(typeOf(value) !== 'undefined' &&
    typeof value !== 'undefined' &&
    value !== undefined, value, payload);

const isNot = Object.freeze({
    boolean: isNotBoolean,
    defined: isNotDefined,
    function: isNotFunction,
    null: isNotNull,
    number: isNotNumber,
    string: isNotString,
    undefined: isNotUndefined
});

const is = Object.freeze({
    array: isArray,
    bigint: isBigInt,
    boolean: isBoolean,
    booleanObject: isBooleanObject,
    booleanType: isBooleanType,
    class: isClass,
    date: isDate,
    defined: isDefined,
    false: isFalse,
    function: isFunction,
    instance: isInstance,
    key: isKey,
    not: isNot,
    null: isNull,
    number: isNumber,
    numberBetween: isNumberBetween,
    numberObject: isNumberObject,
    numberType: isNumberType,
    object: isObject,
    objectKey: isObjectKey,
    objectKeyIn: isObjectKeyIn,
    objectKeys: isObjectKeys,
    objectKeysIn: isObjectKeysIn,
    objectSomeKeys: isObjectSomeKeys,
    primitive: isPrimitive,
    regexp: isRegExp,
    string: isString,
    stringIncludes: isStringIncludes,
    stringIncludesSome: isStringIncludesSome,
    stringLength: isStringLength,
    stringLengthBetween: isStringLengthBetween,
    stringObject: isStringObject,
    stringType: isStringType,
    symbol: isSymbol,
    true: isTrue,
    type: isType,
    undefined: isUndefined
});

function isParam(...param) {
    return (target, key, descriptor) => {
        const originalMethod = descriptor.value;
        descriptor.value = function () {
            if (is.array(param) && is.defined(arguments)) {
                param.forEach((name, index) => {
                    if (is.number(index) && index < arguments.length) {
                        if (is.defined(arguments[index])) {
                            switch (name) {
                                case 'number':
                                    if (is.number(arguments[index]) === false) {
                                        arguments[index] = undefined;
                                    }
                                    break;
                                case 'object':
                                    if (is.object(arguments[index]) === false) {
                                        arguments[index] = undefined;
                                    }
                                    break;
                                case 'string':
                                    if (is.string(arguments[index]) === false) {
                                        arguments[index] = undefined;
                                    }
                                    break;
                            }
                        }
                    }
                });
            }
            const result = originalMethod.apply(this, arguments);
            return result;
        };
        return descriptor;
    };
}

const guardStringLength = (value, length, callback, payload) => isStringLength(value, length, callback, payload);

const guardStringLengthBetween = (value, min, max, callback, payload) => isStringLengthBetween(value, min, max, callback, payload);

const guardSymbol = (value, callback, payload) => isSymbol(value, callback, payload);

const guardTrue = (value, callback, payload) => isTrue(value, callback, payload);

const guardType = (value, type, callback, payload) => isType(value, type, callback, payload);

const guardUndefined = (value, callback, payload) => isUndefined(value, callback, payload);

const guardIs = Object.freeze({
    array: guardArray,
    bigint: guardBigInt,
    boolean: guardBoolean,
    class: guardClass,
    date: guardDate,
    defined: guardDefined,
    false: guardFalse,
    function: guardFunction,
    instance: guardInstance,
    key: guardKey,
    null: guardNull,
    number: guardNumber,
    numberBetween: guardNumberBetween,
    object: guardObject,
    objectKey: guardObjectKey,
    objectKeyIn: guardObjectKeyIn,
    objectKeys: guardObjectKeys,
    objectKeysIn: guardObjectKeysIn,
    objectSomeKeys: guardObjectSomeKeys,
    primitive: guardPrimitive,
    regexp: guardRegExp,
    string: guardString,
    stringIncludes: guardStringIncludes,
    stringIncludesSome: guardStringIncludesSome,
    stringLength: guardStringLength,
    stringLengthBetween: guardStringLengthBetween,
    symbol: guardSymbol,
    true: guardTrue,
    type: guardType,
    undefined: guardUndefined
});

const guard = Object.freeze(Object.assign({}, guardIs));

const type = Object.freeze({
    are,
    is,
    guard
});

const RECOGNIZE_INSTANCES = [
    Array,
    ArrayBuffer,
    Boolean,
    DataView,
    Date,
    Error,
    EvalError,
    Float32Array,
    Float64Array,
    Function,
    Int16Array,
    Int32Array,
    Int8Array,
    Map,
    Number,
    Object,
    Promise,
    RangeError,
    ReferenceError,
    RegExp,
    Set,
    Storage,
    String,
    SyntaxError,
    TypeError,
    URIError,
    Uint16Array,
    Uint32Array,
    Uint8Array,
    Uint8ClampedArray,
    WeakMap,
    WeakSet
];

const recognizeValue = (value, onlyTrue = true, instances = []) => {
    const ofRecognized = {
        'Array.isArray': Array.isArray(value),
        isClass: is.class(value),
        isFunction: is.function(value),
        'Number.isInteger': Number.isInteger(value),
        'Number.isFinite': Number.isFinite(value),
        'Number.isNaN': Number.isNaN(value),
        'Number.isSafeInteger': Number.isSafeInteger(value),
        typeOf: typeOf(value),
        typeof: typeof value,
    };
    try {
        Object.assign(ofRecognized, { isFinite: isFinite(value) });
    }
    catch (e) { }
    try {
        Object.assign(ofRecognized, { isNaN: isNaN(value) });
    }
    catch (e) { }
    RECOGNIZE_INSTANCES.concat(instances).forEach((instance) => (Object.assign(ofRecognized, { [instance.name]: value instanceof instance })));
    if (is.true(onlyTrue)) {
        Object.keys(ofRecognized).filter((type) => is.false(ofRecognized[type])
            ? delete ofRecognized[type]
            : true);
    }
    return ofRecognized;
};

export { are, areBigInt, areBoolean, areDate, areDefined, areFalse, areNull, areNumber, areRegExp, areString, areSymbol, areTrue, areUndefined, guard, guardArray, guardBigInt, guardBoolean, guardClass, guardDate, guardDefined, guardFalse, guardFunction, guardInstance, guardKey, guardNull, guardNumber, guardNumberBetween, guardObject, guardObjectKey, guardObjectKeyIn, guardObjectKeys, guardObjectKeysIn, guardObjectSomeKeys, guardPrimitive, guardRegExp, guardString, guardStringIncludes, guardStringIncludesSome, guardStringLength, guardStringLengthBetween, guardSymbol, guardTrue, guardType, guardUndefined, is, isArray, isBigInt, isBoolean, isBooleanObject, isBooleanType, isClass, isDate, isDefined, isFalse, isFunction, isInstance, isKey, isNot, isNotBoolean, isNotDefined, isNotFunction, isNotNull, isNotNumber, isNotString, isNotUndefined, isNull, isNumber, isNumberBetween, isNumberObject, isNumberType, isObject, isObjectKey, isObjectKeyIn, isObjectKeys, isObjectKeysIn, isObjectSomeKeys, isParam, isPrimitive, isRegExp, isString, isStringIncludes, isStringIncludesSome, isStringLength, isStringLengthBetween, isStringObject, isStringType, isSymbol, isTrue, isType, isUndefined, recognizeValue, type, typeOf };
